/***************************************************************************
 *   Copyright (C) 2004-2005 by Christoph Hormann <chris_hormann@gmx.de>   *
 *                                                                         *
 *   Permission is hereby granted, free of charge, to any person obtaining *
 *   a copy of this software and associated documentation files (the       *
 *   "Software"), to deal in the Software without restriction, including   *
 *   without limitation the rights to use, copy, modify, merge, publish,   *
 *   distribute, sublicense, and/or sell copies of the Software, and to    *
 *   permit persons to whom the Software is furnished to do so, subject to *
 *   the following conditions:                                             *
 *                                                                         *
 *   The above copyright notice and this permission notice shall be        *
 *   included in all copies or substantial portions of the Software.       *
 *                                                                         *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,       *
 *   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF    *
 *   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. *
 *   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR     *
 *   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, *
 *   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR *
 *   OTHER DEALINGS IN THE SOFTWARE.                                       *
 ***************************************************************************/

#include <kinstance.h>
#include <kdebug.h>
#include <klocale.h>
#include <kconfig.h>

#include <kstandarddirs.h>
#include <kurl.h>

#include <qstring.h>
#include <qcstring.h>
#include <qsocket.h>
#include <qdatetime.h>
#include <qfile.h>
#include <qfileinfo.h>
#include <qtextstream.h>
#include <qdatastream.h>
#include <qregexp.h>

#include <iostream>
#include <cstdlib>
#include <cmath>
#include <unistd.h>

#include "pov.moc"

using namespace KIO;

extern "C"
{
	int KDE_EXPORT kdemain(int argc, char **argv);
}

/* --------------------------------------------------------------------------- */

kio_povProtocol::kio_povProtocol(const QCString &pool_socket, const QCString &app_socket)
    : SlaveBase("kio_pov", pool_socket, app_socket)
{
	kdDebug(DEBUG_AREA) << "kio_povProtocol::kio_povProtocol()" << endl;
	ConvertKProcess = 0;
}

/* --------------------------------------------------------------------------- */

kio_povProtocol::~kio_povProtocol()
{
	if (ConvertKProcess != 0)
	{
		delete ConvertKProcess;
		ConvertKProcess = 0;
	}
	kdDebug(DEBUG_AREA) << "kio_povProtocol::~kio_povProtocol()" << endl;
}

/* --------------------------------------------------------------------------- */

/**
get a file:

  - if called for an existing file this is rendered or information is displayed
    depending on the options
  - if file does not exist general info mode is enabled
 */
void kio_povProtocol::get(const KURL& url )
{
	processedBytes = 0;

	kdDebug(DEBUG_AREA) << "kio_pov::get(const KURL& url)" << endl ;
	kdDebug(DEBUG_AREA) << "url: " << url.prettyURL() << endl;

	QString surl = url.path().local8Bit();
	if (surl.at(0) == '/') {
		if (KStandardDirs::exists(surl)) 
		{
			ProcessFile(url);  // file -> file mode
		}
		else // no file -> info mode
		{
			mimeType("text/html");
			data(GenerateInfo(url).local8Bit());
		}
	}
	else // no file -> info mode
	{
		mimeType("text/html");
		data(GenerateInfo(url).local8Bit());
	}

	data(QByteArray()); // empty array means we're done sending the data
	finished();
}

/* --------------------------------------------------------------------------- */

/**
determine mime type:
 */

void kio_povProtocol::mimetype(const KURL & url)
{
	QString surl = url.path().local8Bit();
	if (surl.at(0) == '/') {
		if (KStandardDirs::exists(surl)) 
		{
			mimeType("text/html"); // FIXME determine real type (tricky)
		}
		else // no file -> info mode
		{
			mimeType("text/html");
		}
	}
	else // no file -> info mode
	{
		mimeType("text/html");
	}
	
	finished();
}

/* --------------------------------------------------------------------------- */

/**
load a template for the HTML info pages

  @param filename name of the template to lead
  @param buffer the buffer to load the template to

  @return true if loading was successful false otherwise

 */
 
bool kio_povProtocol::loadTemplate(const QString& filename, QString& buffer)
{
	QFile	f(locate("data", QString::fromLatin1("kio_pov/template/")+filename));
	if (f.exists() && f.open(IO_ReadOnly))
	{
		QTextStream	t(&f);
		buffer = t.read();
		return true;
	}
	else
	{
		buffer = QString::null;
		return false;
	}
}

/* --------------------------------------------------------------------------- */

/**
format a POV-Ray log file for display in HTML.

  @param log QString with the log text
  @param Escape if true HTML characters are replaced
                otherwise not (in case there is already HTML in the log)

  @return QString containing the formatted log

 */

QString kio_povProtocol::HTMLLog(const QString & log, const bool Escape)
{
	QString log_text(log);

	if (Escape)
	{
		log_text.replace(QRegExp("&"), "&amp;");
		log_text.replace(QRegExp("<"), "&lt;");
		log_text.replace(QRegExp(">"), "&gt;");
		log_text.replace(QRegExp("\""), "&quot;");
		log_text.replace(QRegExp("-----------------[-]+\n"), "<hr>");
	}

	return QString(
			"<div style=\"margin: 0px; padding: 3px; color: #FFFFFF; background-color: #324EAF;\">"
			"<div style=\"border: 1px inset #6080E0; padding: 3px;\">"
			"<pre style=\"font-family: monospace; padding: 3pt; white-space: pre;\">"
			"%1"
			"</pre></div></div>").arg(log_text);
}

/* --------------------------------------------------------------------------- */

/**
generate a HTML table with two columns using
information from two QStringLists

  @param col1 QStringLists containing the first column table elements
  @param col2 QStringLists containing the second column table elements

  @return QString containing the table

 */

QString kio_povProtocol::HTMLInfoTable(const QStringList & col1, const QStringList & col2)
{
	QString Table(
			"<div>"
			"<table cellspacing=2 style=\"padding: 0; margin: 10px; border-top: 1px solid #4A81D5; border-bottom: 1px solid #4A81D5;\">");

	QStringList::ConstIterator i = col1.begin();
	QStringList::ConstIterator j = col2.begin();

	for (;(i != col1.end()) && (j != col1.end());++i, ++j)
	{
		Table += "<tr>";
		Table += "<td style=\"background-color:#DDDDFF;\">";
		Table += *i;
		Table += "</td>";
		Table += "<td align=\"right\">";
		Table += *j;
		Table += "</td>";
		Table += "</tr>";
	}

	Table += "</table></div>";

	return Table;
}

/* --------------------------------------------------------------------------- */

/**
generates a text message in plain HTML
to signal something went wrong

  @param title title of the HTML page
  @param message the message to display

 */

void kio_povProtocol::HTMLMessage(const QString & title, const QString & message)
{
	QString html("<html><head><title>%1</title></head><body>%1</body></html>");
	mimeType("text/html");
	data(html.arg(title).arg(message).local8Bit());
}

/* --------------------------------------------------------------------------- */

/**
search the Library_Path to find the position of a certain include file

  @param Povray POVExecutable to check the Library_Path settings for
  @param File QString containing the include file name
  @param Scene_Dir QString containing the scene path -
                   this dir is checked before the Library_Path dirs

  @return The full path of the include file

 */

QString kio_povProtocol::Find_POV_Include(POVExecutable & Povray, const QString& File, const QString& Scene_Dir)
{
	bool Found=false;

	if (Povray.exists())
	{
		QString path = Scene_Dir + "/" + File;
		if (!KStandardDirs::exists(path))
		{
			const QStringList LibPaths = Povray.Library_Paths();
			for ( QStringList::ConstIterator i = LibPaths.begin(); i != LibPaths.end(); ++i )
			{
				path = QString(*i + "/" + File);
				if (KStandardDirs::exists(path))
				{
					Found=true;
					break;
				}
			}
		}
		else Found=true;
		if (Found) return path;
		else return QString::null;
	}
	else return QString::null;
}

/* --------------------------------------------------------------------------- */

/**
generate a general info page

  @param url URL containing possible parameters (currently unused)

  @return QString containing the info page

 */
 
QString kio_povProtocol::GenerateInfo(const KURL & url)
{
	QString	content;

	if (!loadTemplate(QString::fromLatin1("kio_pov.template"), content))
	{
		error(KIO::ERR_INTERNAL, i18n("Unable to load template %1").arg("kio_pov.template"));
		return content;
	}

	QString HomeDir;
	char *s;
	if ((s = getenv("HOME"))) HomeDir = QString(s);
	else
	{
		error(KIO::ERR_INTERNAL, i18n("Fatal error - unable to read HOME environment variable"));
		return content;
	}

	// ----------------- POV-Ray -----------------------------
	
	infoMessage(i18n("Checking POV-Ray configuration..."));

	QString POV_Install("<h2>"+i18n("POV-Ray installation:")+"</h2>");
	
	// test run POV-Ray
	POVExecutable Povray("povray");
	if (Povray.exists())
	{
		POV_Install +=
				QString(
					"<div>"+
					i18n(
						"The POV-Ray executable started by running %1 is %1. "
						"It generates the following version information: "
						"%1 %1"
					)+
					"</div>"
				)
				.arg("<code>megapov</code>")
				.arg("<code>"+Povray.path()+"<code>")
				.arg("<pre>"+Povray.Version_Information()+"</pre>")
				.arg("<p><strong>"+Povray.Compile_Information()+"</strong></p>");
	
	}
	else POV_Install += "<div><strong>"+i18n("no POV-Ray executable found</strong>")+"</div>";

	
	// determine if POV-Ray is completely installed
	{
		QStringList Col1;
		QStringList Col2;
	
		Col1 += i18n("POV-Ray executable");
		if (Povray.exists())
		{
			Col2 += Povray.path();
			Col1 += i18n("POV-Ray version");
			Col2 += Povray.version_string();
		}
		else
			Col2 += "<strong>not found!</strong>";
	
		Col1 += i18n("standard include files and sample scenes");
		QString dir = "/usr/local/share/" + Povray.name() + "-" + Povray.version_string() + "/include/functions.inc";
		if (KStandardDirs::exists(dir))
			Col2 += "<a href=\"file:///usr/local/share/" + Povray.name() + "-" + Povray.version_string() + "/\">exist</a>";
		else
			Col2 += "<strong>"+i18n("do not exist!")+"</strong>";
	
		Col1 += i18n("POV-Ray documentation");
		dir = "/usr/local/share/doc/" + Povray.name() + "-" + Povray.version_string() + "/html/index.html";
		if (KStandardDirs::exists(dir))
			Col2 += "<a href=\"file://" + dir + "\">exists</a>";
		else
			Col2 += "<strong>"+i18n("does not exist!")+"</strong>";
	
		POV_Install += HTMLInfoTable(Col1, Col2);
	}


	QString POV_Conf("<h2>"+i18n("POV-Ray configuration:")+"</h2>");

	POV_Conf +=
			"<p>"+
			i18n("these are the files POV-Ray reads the default render options and IO-Restrictions configuration from:")+
			"</p>";
	
	{
		QStringList Col1;
		QStringList Col2;

		QString POVini = Povray.ini();

		Col1 += i18n("main POV-Ray ini file");
		Col2 += "<a href=\"file://" + POVini + "\">" + POVini + "</a>";
	
		Col1 += i18n("system IO-restrictions configuration file:");
		QString conf = HomeDir + "/." + Povray.name() + "/" + Povray.version_string() + "/povray.conf";
		if (KStandardDirs::exists(conf))
			Col2 += "<a href=\"file://" + conf + "\">exists</a>";
		else
			Col2 += i18n("does not exist");

		Col1 += i18n("user IO-restrictions configuration file:");
		conf = "/usr/local/etc/" + Povray.name() + "/" + Povray.version_string() + "/povray.conf";
		if (KStandardDirs::exists(conf))
			Col2 += "<a href=\"file://" + conf + "\">exists</a>";
		else
			Col2 += i18n("does not exist");
		
		POV_Conf += HTMLInfoTable(Col1, Col2);
	}

	if (Povray.Library_Paths().count()>0)
	{
		POV_Conf += QString(
				"<div><h3>"+
				i18n("Library_Path settings:")+
				"</h3><p>"+
				i18n("these are the directories POV-Ray searches for files when parsing a scene:")+
				"</p><div style=\"margin_left: 10px;\"><pre>%1</pre></div></div>"
											 ).arg(Povray.Library_Paths().join("\n"));
	}

	
	infoMessage(i18n("Checking MegaPOV configuration..."));

	// ----------------- MegaPOV -----------------------------
	QString MP_Install("<h2>"+i18n("MegaPOV installation:")+"</h2>");
	
	// test run MegaPOV
	POVExecutable Megapov("megapov");
	if (Megapov.exists())
	{
		MP_Install +=
				QString(
					"<div>"+
					i18n(
						"The POV-Ray executable started by running %1 is %1. "
						"It generates the following version information: "
						"%1 %1"
					)+
					"</div>"
				)
				.arg("<code>megapov</code>")
				.arg("<code>"+Megapov.path()+"<code>")
				.arg("<pre>"+Megapov.Version_Information()+"</pre>")
				.arg("<p><strong>"+Megapov.Compile_Information()+"</strong></p>");
	
	}
	else MP_Install += "<div><strong>"+i18n("no MegaPOV executable found")+"</strong></div>";

	// determine if MegaPOV is completely installed
	{
		QStringList Col1;
		QStringList Col2;
	
		Col1 += i18n("MegaPOV executable");
		if (Megapov.exists())
		{
			Col2 += Megapov.path();
			Col1 += i18n("MegaPOV version");
			Col2 += Megapov.version_string();
		}
		else
			Col2 += "<strong>"+i18n("not found!")+"</strong>";
	
		Col1 += i18n("standard include files and sample scenes");
		QString dir = "/usr/local/share/" + Megapov.name() + "-" + Megapov.version_string() + "/include/mp_functions.inc";
		if (KStandardDirs::exists(dir))
			Col2 += "<a href=\"file:///usr/local/share/" + Megapov.name() + "-" + Megapov.version_string() + "/\">exist</a>";
		else
			Col2 += "<strong>"+i18n("does not exist!")+"</strong>";
	
		Col1 += i18n("MegaPOV documentation");
		dir = "/usr/local/share/doc/" + Megapov.name() + "-" + Megapov.version_string() + "/html/index.html";
		if (KStandardDirs::exists(dir))
			Col2 += "<a href=\"file://" + dir + "\">exists</a>";
		else
			Col2 += "<strong>"+i18n("does not exist!")+"</strong>";
	
		MP_Install += HTMLInfoTable(Col1, Col2);
	}

	QString MP_Conf("<h2>"+i18n("MegaPOV configuration:")+"</h2>");

	MP_Conf +=
			"<p>"+
			i18n("these are the files MegaPOV reads the default render options and IO-Restrictions configuration from:")+
			"</p>";

	{
		QStringList Col1;
		QStringList Col2;

		QString POVini = Megapov.ini();

		Col1 += i18n("main MegaPOV ini file");
		Col2 += "<a href=\"file://" + POVini + "\">" + POVini + "</a>";
	
		Col1 += i18n("system IO-restrictions configuration file:");
		QString conf = HomeDir + "/." + Megapov.name() + "/" + Megapov.version_string() + "/povray.conf";
		if (KStandardDirs::exists(conf))
			Col2 += "<a href=\"file://" + conf + "\">exists</a>";
		else
			Col2 += i18n("does not exist");

		Col1 += i18n("user IO-restrictions configuration file:");
		conf = "/usr/local/etc/" + Megapov.name() + "/" + Megapov.version_string() + "/povray.conf";
		if (KStandardDirs::exists(conf))
			Col2 += "<a href=\"file://" + conf + "\">exists</a>";
		else
			Col2 += i18n("does not exist");
		
		MP_Conf += HTMLInfoTable(Col1, Col2);
	}


	if (Megapov.Library_Paths().count()>0)
	{
		MP_Conf += QString(
				"<div><h3>"+
				i18n("Library_Path settings:")+
				"</h3><p>"+
				i18n("these are the directories MegaPOV searches for files when parsing a scene:")+
				"</p><div style=\"margin_left: 10px;\"><pre>%1</pre></div></div>"
				).arg(Megapov.Library_Paths().join("\n"));
	}

	QString logo = QString::fromLatin1("file:");
	logo.append(locate("data", QString::fromLatin1("kio_pov/template/logo_alpha.png")));

	content = content
			.arg(i18n("POV-Ray information"))
			.arg(logo)
			.arg(i18n("POV-Ray information"))
			.arg(i18n("POV-Ray information"))
			.arg("<p>"+i18n("This is a summary of information about the POV-Ray installations on this machine")+"</p>" +
			POV_Install + POV_Conf + MP_Install + MP_Conf)
			.arg(KIO_POV_NAME " version " KIO_POV_VERSION);

	return content;
}

/* --------------------------------------------------------------------------- */

/**
renders a scene as still with certain options.

  @param output_data the array the image/video data is returned with.
                     the image data is also written to the cache file
                     animation are generated as file and loaded
  @param scene_info SceneInfo struct containing all neccessary parameters for the render.

  @return

    0   if the render worked without problems.
    1   if there were problems writing the results to the cache files (but the render worked).
    2   if there was a problem starting the render.
    >=3 if there were problems reported by POV-Ray.

 */

int kio_povProtocol::RenderStill(const QByteArray & output_data, POVScene & Scene)
{
	int res = 0;

	POVExecutable Povray(Scene.povray_executable());

	if (!Povray.exists())
	{
		error(ERR_CANNOT_LAUNCH_PROCESS,
					i18n("Could not render the scene because the required %1 executable could be found.").arg(Scene.povray_executable()));
		return 2;
	}

	chdir (Scene.render_directory().ascii());

	PovrayKProcess = new KProcess();

	ErrorStream = QString::null;
	processedBytes = 0;

	Povray_ImageStream = new QDataStream(output_data, IO_WriteOnly);
	*PovrayKProcess << Scene.povray_executable() << Scene.ini_file() << "-o-" << "+fn" << "-p";

	//data is now sent directly from the slot
	connect(PovrayKProcess, SIGNAL(receivedStdout(KProcess *, char *, int)),
					this, SLOT(slotSetDataStdOutput(KProcess *, char *, int)));
	connect(PovrayKProcess, SIGNAL(receivedStderr(KProcess *, char *, int)),
					this, SLOT(slotGetStdError(KProcess *, char *, int)));

	PovrayKProcess->start(KProcess::Block, KProcess::All);

	if (!PovrayKProcess->normalExit()) {
		error(ERR_CANNOT_LAUNCH_PROCESS,
					i18n("There was an error with %1 - please ensure it is correctly installed").arg(Scene.povray_executable()));
		res = 2;
	}

	if (res<2)
	{
		// now write image data to cache file
		if (processedBytes>0)
		{
			QFile img_file( Scene.img_file() );
			if ( img_file.open( IO_WriteOnly ) )
			{
				//infoMessage(i18n(QString("write "+img_name)));
				QDataStream img_stream( &img_file );
				img_stream.writeRawBytes(output_data, output_data.size());
				img_file.close();
				GenerateThumbnail(Scene);
			}
			else
			{
				infoMessage(i18n(QString("Could not write to file %1").arg(Scene.img_file())));
				res = 1;
			}
		}
		else res = 2;
	
		// and save log
		QFile log_file( Scene.log_file() );
		if ( log_file.open( IO_WriteOnly ) )
		{
			QTextStream log_stream( &log_file );
			log_stream << ErrorStream;
			log_file.close();
		}
		else
		{
			infoMessage(i18n("Could not write to file %1").arg(Scene.log_file()));
		}
	}

	QString Error_Message;
	QString Error_Text;

	int res_log = AnalyzeLog(Error_Message, Error_Text);
	if (res_log > 0) res = res_log;
	
	if (res>=3) FormatError(Scene, Error_Message, Error_Text);

	//clean up
	delete PovrayKProcess; PovrayKProcess = 0;
	delete Povray_ImageStream; Povray_ImageStream = 0;

	return res;
}

/* --------------------------------------------------------------------------- */

/**
renders a scene as animation with certain options.

  @param output_data the array the image/video data is returned with.
                     the image data is also written to the cache file
                     animation are generated as file and loaded
  @param scene_info SceneInfo struct containing all neccessary parameters for the render.

  @return

    0   if the render worked without problems.
    1   if there was a problem assembling the animation - we return the first frame image
    2   if there was a problem starting the render.
    >=3 if there were problems reported by POV-Ray.

 */

int kio_povProtocol::RenderAnim(const QByteArray & output_data, POVScene & Scene)
{
	int res = 0;

	POVExecutable Povray(Scene.povray_executable());

	if (!Povray.exists())
	{
		error(ERR_CANNOT_LAUNCH_PROCESS,
					i18n("Could not render the scene because the required %1 executable could be found.").arg(Scene.povray_executable()));
		return 2;
	}

	chdir (Scene.render_directory().ascii());

	PovrayKProcess = new KProcess();

	ErrorStream = QString::null;
	processedBytes = 0;

	QString Out_Option = "-o" + Scene.img_file();

	*PovrayKProcess << Scene.povray_executable() << Scene.ini_file() << Out_Option << "+fn" << "-p";

	connect(PovrayKProcess, SIGNAL(receivedStderr(KProcess *, char *, int)),
					this, SLOT(slotGetStdErrorAnim(KProcess *, char *, int)));

	PovrayKProcess->start(KProcess::Block, KProcess::All);

	if (!PovrayKProcess->normalExit()) {
		error(ERR_CANNOT_LAUNCH_PROCESS,
					i18n("There was an error with %1 - please ensure it is correctly installed").arg(Scene.povray_executable()));
		res = 2;
	}

	if (res<2)
	{
		// save log
		QFile log_file( Scene.log_file() );
		if ( log_file.open( IO_WriteOnly ) )
		{
			QTextStream log_stream( &log_file );
			log_stream << ErrorStream;
			log_file.close();
		}
		else
		{
			infoMessage(i18n("Could not write to file %1").arg(Scene.log_file()));
		}
	}

	QString Error_Message;
	QString Error_Text;

	int res_log = AnalyzeLog(Error_Message, Error_Text);
	if (res_log > 0) res = res_log;

	if (res<2)
	{
		// now assemble the animation and load the video file
		FFMpegErrorStream = QString::null;
		if (KStandardDirs::exists(Scene.img_file_base()+"00001.png")) res = GenerateMpeg(Scene, output_data, 5);
		else if (KStandardDirs::exists(Scene.img_file_base()+"0001.png")) res = GenerateMpeg(Scene, output_data, 4);
		else if (KStandardDirs::exists(Scene.img_file_base()+"001.png")) res = GenerateMpeg(Scene, output_data, 3);
		else if (KStandardDirs::exists(Scene.img_file_base()+"01.png")) res = GenerateMpeg(Scene, output_data, 2);
		else if (KStandardDirs::exists(Scene.img_file_base()+"1.png")) res = GenerateMpeg(Scene, output_data, 1);
		else res = 2;

		if (res>=2) FormatErrorFFMpeg(Scene, FFMpegErrorStream);
	}
	
	if (res>=3) FormatError(Scene, Error_Message, Error_Text);

	//clean up
	delete PovrayKProcess; PovrayKProcess = 0;

	return res;
}

/* --------------------------------------------------------------------------- */

/**
generates an MPEG video from a rendered animation

  @param scene_info SceneInfo struct for the scene to generate the animation for
  @param output_data the array the video data is returned with.
  @param digits number of digits of frame index in file names

  @return

    0   animation successfully assembled
    1   error, no animation but still image of first frame
    2   error, no animation, no still

 */

int kio_povProtocol::GenerateMpeg(POVScene & Scene, const QByteArray & output_data, const int digits)
{
	int res = 2;
	if (KStandardDirs::findExe("ffmpeg") != QString::null)
	{
		infoMessage(i18n("assembling animation using ffmpeg"));
		FFMpegKProcess = new KProcess();

		QString in_file = Scene.img_file_base() + "%" + QString::number(digits) + "d.png";
		*FFMpegKProcess << "ffmpeg" << "-r" << QString::number(Scene.framerate()) << "-b" << QString::number(Scene.bitrate()) << "-i" << in_file << "-y" << Scene.anim_file();

		connect(FFMpegKProcess, SIGNAL(receivedStderr(KProcess *, char *, int)),
						this, SLOT(slotMpegStdOut(KProcess *, char *, int)));
		connect(FFMpegKProcess, SIGNAL(receivedStdout(KProcess *, char *, int)),
						this, SLOT(slotMpegStdOut(KProcess *, char *, int)));

		FFMpegKProcess->start(KProcess::Block, KProcess::All);

		if (FFMpegKProcess->normalExit())
		{
			/*
			// save log
			QFile log_file( Scene.img_file_base()+"_ffmpeg.log" );
			if ( log_file.open( IO_WriteOnly ) )
			{
				QTextStream log_stream( &log_file );
				log_stream << FFMpegErrorStream;
				log_file.close();
			}
			else
			{
				infoMessage(i18n("Could not write to file %1").arg(Scene.img_file_base()+"_ffmpeg.log"));
			}
			*/
			if (KStandardDirs::exists(Scene.anim_file()))
			{
				QFile anim_file( Scene.anim_file() );
				if ( anim_file.open( IO_ReadOnly ) )
				{
					QDataStream anim_stream( &anim_file );
					QDataStream anim_stream_internal(output_data, IO_WriteOnly);
					char * const anim_data = new char[anim_file.size()];
					anim_stream.readRawBytes(anim_data, anim_file.size());
					anim_stream_internal.writeRawBytes(anim_data, anim_file.size());
					delete anim_data;
					infoMessage(i18n("video generated (%1 bytes)").arg(QString::number(anim_file.size())));
					anim_file.close();
					res = 0;
				}
				else
				{
					error(ERR_CANNOT_OPEN_FOR_READING,
								i18n("file %1").arg(Scene.anim_file()));
				}

				// remove the individual frame images
				for (int i=1; i<=pow(10,digits-1); i++)
				{
					QString filename = QString::number(i);
					for (int j=filename.length(); j<digits; j++) filename.prepend('0');
					filename.prepend(Scene.img_file_base());
					filename.append(".png");
					if (KStandardDirs::exists(filename))
						unlink(filename.ascii());
				}
			}
		}
		
		delete FFMpegKProcess; FFMpegKProcess = 0;
	}
	else
	{
		infoMessage(i18n("ffmpeg not found - returning first frame of animation"));
		QString filename = Scene.img_file_base();
		for (int i=1; i < digits; i++) filename.append('0');
		filename.append("1.png");
		QFile img_file( filename );
		if ( img_file.open( IO_ReadOnly ) )
		{
			QDataStream img_stream( &img_file );
			QDataStream img_stream_internal(output_data, IO_WriteOnly);
			char * const img_data = new char[img_file.size()];
			img_stream.readRawBytes(img_data, img_file.size());
			img_stream_internal.writeRawBytes(img_data, img_file.size());
			delete img_data;
			img_file.close();
			res = 1;
		}
		else
		{
			error(ERR_CANNOT_OPEN_FOR_READING, i18n("file %1").arg(filename));
		}
	}

	return res;
}

/* --------------------------------------------------------------------------- */

/**
generates a thumbnail of the render using ImageMagick convert.

  @param scene_info SceneInfo struct for the scene to generate the thumbnail for

 */

void kio_povProtocol::GenerateThumbnail(POVScene & Scene)
{
	if (KStandardDirs::findExe("convert") != QString::null)
		if (Scene.render_width()>KIO_POV_THUMBNAIL_WIDTH)
		{
			if (ConvertKProcess != 0)
			{
				delete ConvertKProcess;
				ConvertKProcess = 0;
			}
			ConvertKProcess = new KProcess();

			*ConvertKProcess << "convert" << Scene.img_file() << "-scale" << QString::number(KIO_POV_THUMBNAIL_WIDTH)+"x"+QString::number(KIO_POV_THUMBNAIL_HEIGHT) << Scene.thumbnail_file();

			infoMessage(i18n("Generating Thumbnail"));
			ConvertKProcess->start(KProcess::NotifyOnExit, KProcess::All);
		}
}

/* --------------------------------------------------------------------------- */

/**
formats an error message from POV-Ray for display

  @param scene_info SceneInfo struct containing the file names.
  @param Error_Message error message returned by POV-Ray,
                       formatted by AnalyzeLog()
  @param Error_Text full error text returned by POV-Ray,
                    formatted by AnalyzeLog()

  @return

    true if there were problems formatting the error message
    false otherwise

 */

bool kio_povProtocol::FormatError(POVScene & Scene, const QString & Error_Message, const QString & Error_Text)
{
	QString	content;
	if (!loadTemplate(QString::fromLatin1("kio_pov.template"), content))
	{
		error(KIO::ERR_INTERNAL, i18n("Unable to load template %1").arg("kio_pov.template"));
		return true;
	}
	QString logo = QString::fromLatin1("file:");
	logo.append(locate("data", QString::fromLatin1("kio_pov/template/logo_alpha.png")));

	content = content
				.arg(Scene.pov_file())
				.arg(logo)
				.arg(Scene.filename())
				.arg(Error_Message)
				.arg(QString(
				"<p>"+i18n("An error occured while rendering this scene.  POV-Ray reports:")+"</p>"
				"%1"
										).arg(Error_Text))
				.arg(KIO_POV_NAME " version " KIO_POV_VERSION);
			
		mimeType("text/html");
		data(content.local8Bit());

	return false;
}

/* --------------------------------------------------------------------------- */

/**
formats an error message from FFMpeg for display

  @param scene_info SceneInfo struct containing the file names.
  @param Error_Message error message returned by FFMpeg

  @return

    true if there were problems formatting the error message
    false otherwise

 */

bool kio_povProtocol::FormatErrorFFMpeg(POVScene & Scene, const QString & Error_Message)
{
	QString	content;
	if (!loadTemplate(QString::fromLatin1("kio_pov.template"), content))
	{
		error(KIO::ERR_INTERNAL, i18n("Unable to load template %1").arg("kio_pov.template"));
		return true;
	}
	QString logo = QString::fromLatin1("file:");
	logo.append(locate("data", QString::fromLatin1("kio_pov/template/logo_alpha.png")));

	content = content
			.arg(Scene.pov_file())
			.arg(logo)
			.arg(Scene.filename())
			.arg(i18n("problem generating a video from the animation render"))
			.arg(QString(
			"<p>"+i18n("An error occured while assembling the video.  FFMpeg reports:")+"</p>"
			"%1").arg(HTMLLog(Error_Message, true)))
			.arg(KIO_POV_NAME " version " KIO_POV_VERSION);
			
	mimeType("text/html");
	data(content.local8Bit());

	return false;
}

/* --------------------------------------------------------------------------- */

/**
formats the render log file for display

  @param scene_info SceneInfo struct containing the file names.
  @param Variant 0: plain text
                 1: full HTML formatted log
                 2: only render statistics

  @return

    true if there were problems formatting the log
    false otherwise

 */

bool kio_povProtocol::FormatLog(const POVScene & Scene, const int Variant)
{
	// first read log file
	QString log_text;
	QFile log_file( Scene.log_file() );
	if ( log_file.open( IO_ReadOnly ) )
	{
		QTextStream log_stream( &log_file );
		log_text = log_stream.read();
		log_text.replace('\r', '\n');
		log_file.close();
	}
	else
	{
		infoMessage(i18n("Could not read from file %1").arg(Scene.log_file()));
		log_text = i18n("Could not read log file %1").arg(Scene.log_file());
	}

	QString pov_path=Scene.pov_file();
	
	if (Variant==1) // full HTML formatted log
	{
		QString	content;
		if (!loadTemplate(QString::fromLatin1("kio_pov.template"), content))
		{
			error(KIO::ERR_INTERNAL, i18n("Unable to load template %1").arg("kio_pov.template"));
			return true;
		}
		QString logo = QString::fromLatin1("file:");
		logo.append(locate("data", QString::fromLatin1("kio_pov/template/logo_alpha.png")));

		content = content
				.arg(pov_path)
				.arg(logo)
				.arg(Scene.filename())
				.arg(i18n("render log"))
				.arg(QString(
				"<ul>"
				"<li><a href=\"pov:%1?info\">"+i18n("view scene information")+"</a></li>"
				"<li><a href=\"pov:%1?render\">"+i18n("view the image rendered from this scene")+"</a></li>"
				"</ul>"
				"%1")
				.arg(pov_path)
				.arg(pov_path)
				.arg(HTMLLog(log_text, true)))
				.arg(KIO_POV_NAME " version " KIO_POV_VERSION);
			
		mimeType("text/html");
		data(content.local8Bit());
	}
	else if (Variant==2) // only stats
	{
		QString	content;
		if (!loadTemplate(QString::fromLatin1("kio_pov.template"), content))
		{
			error(KIO::ERR_INTERNAL, i18n("Unable to load template %1").arg("kio_pov.template"));
			return true;
		}
		QString logo = QString::fromLatin1("file:");
		logo.append(locate("data", QString::fromLatin1("kio_pov/template/logo_alpha.png")));

		int start_pos = log_text.find("Render Statistics", 0);
		
		if (start_pos<0) log_text = i18n("no render statstics found in log");

		log_text.remove(0, start_pos);

		content = content
				.arg(pov_path)
				.arg(logo)
				.arg(Scene.filename())
				.arg(i18n("render statistics"))
				.arg(QString(
				"<ul>"
				"<li><a href=\"pov:%1?log\">"+i18n("view full render log")+"</a></li>"
				"<li><a href=\"pov:%1?info\">"+i18n("view scene information")+"</a></li>"
				"<li><a href=\"pov:%1?render\">"+i18n("view the image rendered from this scene")+"</a></li>"
				"</ul>"
				"%1")
				.arg(pov_path)
				.arg(pov_path)
				.arg(pov_path)
				.arg(HTMLLog(log_text, true)))
				.arg(KIO_POV_NAME " version " KIO_POV_VERSION);
		
		mimeType("text/html");
		data(content.local8Bit());
	}
	else // plain text
	{
		mimeType("text/x-log");
		data(QCString( log_text.ascii() ));
	}
	
	return false;
}

/* --------------------------------------------------------------------------- */

/**
analyze the POV-Ray log:
  - determine if the render was successful
  - extract error messages

  @param Error_Message QString the error message returned by POV-Ray
                       is returned in.  When there was no error
                       this is not changed.
  @param Error_Text QString with the formatted full error text
                    returned by POV-Ray.  When there was no error
                    this is not changed.

  @return

    0: render completed without error
    3: unknown error
    4: parse error
    5: 'Failed to render file'-error

 */

int kio_povProtocol::AnalyzeLog(QString & Error_Message, QString & Error_Text)
{
	int pos;

	pos = ErrorStream.findRev( QRegExp("[\\r\\n][ ]*[0-9]+:[0-9]+:[0-9]+[ ]*Done Tracing") );
	if (pos >= 0)
	{
		pos = ErrorStream.find( QRegExp("[\\r\\n][ ]*Total Time:"), pos );
		if (pos >= 0)
		{
			pos++;
			int end_pos = ErrorStream.find( QRegExp("[\\r\\n]"), pos );
			if (end_pos >= 0)
			{
				QString message = ErrorStream.mid(pos,end_pos-pos);
				infoMessage(i18n("POV-Ray finished: %1").arg(message));
			}
		}
		return 0;
	}
	else // no clean exit - search for common errors
	{
		if ((pos = ErrorStream.findRev( "Failed to render file due to error" )) >= 0)
		{
			int end_pos = ErrorStream.find( QRegExp("[\\r\\n]"), pos );
			if (end_pos >= 0)
			{
				Error_Message = ErrorStream.mid(pos,end_pos-pos);
			}
			else Error_Message = "Failed to render file due to error";

			Error_Text = HTMLLog(ErrorStream, true);
			return 5;
		}
		else if ((pos = ErrorStream.findRev( QRegExp("[\\r\\n]Parse Error:[ ]+") )) >= 0)
		{
			pos++;
			int end_pos = ErrorStream.find( QRegExp("[\\r\\n]"), pos );
			if (end_pos >= 0)
			{
				Error_Message = ErrorStream.mid(pos,end_pos-pos);
			}
			else Error_Message = "Parse Error";

			Error_Text = ErrorStream;
			Error_Text.replace(QRegExp("&"), "&amp;");
			Error_Text.replace(QRegExp("<"), "&lt;");
			Error_Text.replace(QRegExp(">"), "&gt;");
			Error_Text.replace(QRegExp("\""), "&quot;");
			Error_Text.replace(QRegExp("-----------------[-]+\n"), "<hr>");

			pos = Error_Text.findRev( QRegExp("[\\r\\n]Parse Error:[ ]+") );
			pos++;
			end_pos = Error_Text.find( QRegExp("[\\r\\n]"), pos );
			Error_Text.insert(end_pos, "</span>");
			Error_Text.insert(pos, "<span style=\"color: #FFFFFF; background-color: #FF3020;\">");

			int pos2 = Error_Text.findRev( QRegExp("[\\r\\n]File:[ ]+"), pos );
			end_pos = Error_Text.find( QRegExp("[ \\r\\n]"), pos2+7 );
			QString error_file = Error_Text.mid(pos2,end_pos-pos2).section( ' ', 1, 1, QString::SectionSkipEmpty );
			Error_Text.insert(end_pos, "</a>");
			Error_Text.insert(pos2+7, "<a href=\"file:" + error_file + "\">");
			
			Error_Text = HTMLLog(Error_Text, false);
			return 4;
		}
		else // other errors
		{
			Error_Message = "Unknown Error";
			Error_Text = HTMLLog(ErrorStream, true);
			return 3;
		}
	}
	return 0;
}

/* --------------------------------------------------------------------------- */

/**
process an existing file according to the options in the URL

  @param url the complete URL with file name and possible options

 */

void kio_povProtocol::ProcessFile(const KURL & url)
{
	POVScene Scene(url.path().local8Bit());
	QString query = url.query().local8Bit();
	
	// ------------------------------------------------------------------------
	// render mode:
	//   if there is no up-to-date cached render
	//   render the scene.
	//   otherwise load the cached image.
	if (query.mid(1, 6) == "render")
	{
		// forced rerender?
		bool force_render = false;
		QString param=url.queryItem("render", 0);
		if (param.left(3).lower() == "new")
			force_render = true;

		// determine render options supposed to be used
		QString render_options=url.queryItem("options", 0).ascii();
		if (render_options!=QString::null)
		{
			// auto options: try to read from scene file header comment
			if (render_options.left(4).lower() == "auto")
				render_options = Scene.options();
		}

		if (!force_render)
		{
			// cache valid - need to compare render options
			// if they have changed force a new render
			if ((Scene.cached() == POVScene::cache_valid) || (Scene.cached() == POVScene::cache_anim))
			{
				if (Scene.ini_options() != render_options)
				{
					//messageBox(Information, "new options: " + render_options);
					force_render = true;
				}
			}
			else force_render = true;
		}
		// forced to use cached render?
		if (force_render)
			if ((param.left(5).lower() == "cache") && (Scene.cached() > POVScene::cache_none))
				force_render = false;
		
		// need to render?
		if (force_render)
		{
			Scene.Prepare();
			if ( Scene.write_options(render_options) )
			{
				QByteArray output_data;
				int render_res;

				if (Scene.isAnimation())
				{
					render_res = RenderAnim(output_data, Scene);
					if (render_res<1) // video file could be generated
					{
						mimeType("video/mpeg");
						data(output_data);
					}
					else if (render_res<2) // no video but first frame image
					{
						mimeType("image/png");
						data(output_data);
					}
				}
				else
				{
					render_res = RenderStill(output_data, Scene);
					if (render_res<2) // no critical error
					{
						mimeType("image/png");
						data(output_data);
					}
				}
			}
			else
			{
				error(ERR_CANNOT_OPEN_FOR_WRITING,
							i18n("ini file %1").arg(Scene.ini_file()));
				return;
			}
		}
		else if ((Scene.cached() == POVScene::cache_anim) || (Scene.cached() == POVScene::cache_old_anim)) // load cached video
		{
			QFile anim_file( Scene.anim_file() );
			if ( anim_file.open( IO_ReadOnly ) )
			{
				mimeType("video/mpeg");
				QDataStream anim_stream( &anim_file );
				char * const anim_data = new char[anim_file.size()];
				anim_stream.readRawBytes(anim_data, anim_file.size());
				QByteArray output_data;
				output_data.setRawData(anim_data, anim_file.size());
				data(output_data);
				output_data.resetRawData(anim_data, anim_file.size());
				delete anim_data;
				anim_file.close();
			}
			else
			{
				error(ERR_CANNOT_OPEN_FOR_READING,
							i18n("file %1").arg(Scene.anim_file()));
				return;
			}
		}
		else // load cached image
		{
			QFile img_file( Scene.img_file() );
			if ( img_file.open( IO_ReadOnly ) )
			{
				mimeType("image/png");
				QDataStream img_stream( &img_file );
				char * const img_data = new char[img_file.size()];
				img_stream.readRawBytes(img_data, img_file.size());
				QByteArray output_data;
				output_data.setRawData(img_data, img_file.size());
				data(output_data);
				output_data.resetRawData(img_data, img_file.size());
				delete img_data;
				img_file.close();
			}
			else
			{
				error(ERR_CANNOT_OPEN_FOR_READING,
							i18n("file %1").arg(Scene.img_file()));
				return;
			}
		}
	}
	// ------------------------------------------------------------------------
	// thumbnail mode:
	//   if there is a cached thumbnail available - return it
	//   otherwise load the full image if not too large.
	//   !!! does not check if the render is up-to-date !!!
	else if (query.mid(1, 9) == "thumbnail")
	{
		if (KStandardDirs::exists(Scene.thumbnail_file()))
		{
			QFile thumb_file( Scene.thumbnail_file() );
			if ( thumb_file.open( IO_ReadOnly ) )
			{
				mimeType("image/png");
				QDataStream thumb_stream( &thumb_file );
				char * const thumb_data = new char[thumb_file.size()];
				thumb_stream.readRawBytes(thumb_data, thumb_file.size());
				QByteArray output_image;
				output_image.setRawData(thumb_data, thumb_file.size());
				data(output_image);
				output_image.resetRawData(thumb_data, thumb_file.size());
				delete thumb_data;
				thumb_file.close();
			}
			else
			{
				error(ERR_CANNOT_OPEN_FOR_READING,
							i18n("file %1").arg(Scene.thumbnail_file()));
				return;
			}
		}
		else if (Scene.render_width()<=KIO_POV_MAX_THUMBNAIL_SIZE)
		{
			if (KStandardDirs::exists(Scene.img_file()))
			{
				QFile thumb_file( Scene.img_file() );
				if ( thumb_file.open( IO_ReadOnly ) )
				{
					mimeType("image/png");
					QDataStream thumb_stream( &thumb_file );
					char * const thumb_data = new char[thumb_file.size()];
					thumb_stream.readRawBytes(thumb_data, thumb_file.size());
					QByteArray output_image;
					output_image.setRawData(thumb_data, thumb_file.size());
					data(output_image);
					output_image.resetRawData(thumb_data, thumb_file.size());
					delete thumb_data;
					thumb_file.close();
				}
				else
				{
					error(ERR_CANNOT_OPEN_FOR_READING,
								i18n("file %1").arg(Scene.img_file()));
					return;
				}
			}
		}
		else
			HTMLMessage(
					Scene.pov_file(),
					QString(i18n("No cached render results found for scene %1. Call first in render mode to generate thumbnail.").arg(Scene.pov_file()))
			);
	}
	// ------------------------------------------------------------------------
	// log mode: 
	//   if there is a cached render display its log file
	//   otherwise issue an error, do not render the scene just for the log
	else if (query.mid(1, 3) == "log")
	{
		// cached render exists?
		if (Scene.cached() > POVScene::cache_none)
		{
			QString render_param=url.queryItem("log", 0);
			if (render_param==QString::null)
				FormatLog(Scene, 1);
			else if (render_param.lower()=="raw")
				FormatLog(Scene, 0);
			else if (render_param.lower()=="stats")
				FormatLog(Scene, 2);
			else
				FormatLog(Scene, 1);
		}
		else
		{
			HTMLMessage(
					Scene.pov_file(),
					QString(i18n("No cached render results found for scene %1. Call first in render mode to generate log.").arg(Scene.pov_file()))
			);
		}
	}
	// ------------------------------------------------------------------------
	// info mode:
	//   generate info page for the scene
	//   this is the default action (if there is no query)
	else //if (query.mid(1, 4) == "info")
	{
		infoMessage(i18n("Analyzing scene..."));

		POVExecutable Povray(Scene.povray_executable());

		QString STest = Povray.version_string();
		
		QString Scene_Menu(
				"<p>"+
				i18n(
					"This is the information page about the POV-Ray scene/include file %1."
					"The following actions are available for this scene:"
				).arg("<code>" + Scene.filename() + "</code>")+
				"</p>"
		);

		// cached render exists?
		switch (Scene.cached())
		{
			case POVScene::cache_valid:
			{
				Scene_Menu +=
					"<ul>"
						"<li><a href=\"pov:" + Scene.pov_file() + "?render=cache\">"+i18n("view the image rendered from this scene (options used: %1)").arg(Scene.ini_options())+"</a></li>"
						"<li><a href=\"pov:" + Scene.pov_file() + "?log\">"+i18n("view render log")+"</a></li>"
						"<li><a href=\"pov:" + Scene.pov_file() + "?render=new\">"+i18n("rerender this scene")+"</a></li>";

				if (Scene.options().length() > 0)
					Scene_Menu +="<li><a href=\"pov:" + Scene.pov_file() + "?render=new&options=auto\">"+i18n("rerender this scene with recommended options")+"</a></li>";

				Scene_Menu += "</ul>";

				if ((KStandardDirs::exists(Scene.thumbnail_file())) || (Scene.render_width()<=KIO_POV_MAX_THUMBNAIL_SIZE))
					Scene_Menu += "<p><a href=\"pov:" + Scene.pov_file() + "?render\"><img src=\"pov:" + Scene.pov_file() + "?thumbnail\" border=0></a></p>";
			}
			break;
			case POVScene::cache_anim:
			{
				Scene_Menu +=
						"<ul>"
						"<li><a href=\"pov:" + Scene.pov_file() + "?render=cache\">"+i18n("view the animation rendered from this scene (options used: %1)").arg(Scene.ini_options())+"</a></li>"
						"<li><a href=\"pov:" + Scene.pov_file() + "?log\">"+i18n("view render log")+"</a></li>"
						"<li><a href=\"pov:" + Scene.pov_file() + "?render=new\">"+i18n("rerender this scene")+"</a></li>";

				if (Scene.options().length() > 0)
					Scene_Menu +="<li><a href=\"pov:" + Scene.pov_file() + "?render=new&options=auto\">"+i18n("rerender this scene with recommended options")+"</a></li>";

				Scene_Menu += "</ul>";
			}
			break;
			default:
			{
				Scene_Menu +=
					"<ul>"
						"<li><a href=\"pov:" + Scene.pov_file() + "?render\">"+i18n("render this scene")+"</a></li>";

				if (Scene.options().length() > 0)
					Scene_Menu += "<li><a href=\"pov:" + Scene.pov_file() + "?render&options=auto\">"+i18n("render this scene with recommended options")+"</a></li>";

				Scene_Menu += "</ul>";
				}
		}

		QString	content;
		if (!loadTemplate(QString::fromLatin1("kio_pov.template"), content))
		{
			error(KIO::ERR_INTERNAL, i18n("Unable to load template %1").arg("kio_pov.template"));
			return;
		}
		QString logo = QString::fromLatin1("file:");
		logo.append(locate("data", QString::fromLatin1("kio_pov/template/logo_alpha.png")));

		// generate include file list
		QString Includes;
		QString Inc_Path;
		const QStringList Inc_List = Scene.includes();
		for ( QStringList::ConstIterator i = Inc_List.begin(); i != Inc_List.end(); ++i )
		{
			Inc_Path = Find_POV_Include(Povray, *i, Scene.scene_directory());
			if (Inc_Path != QString::null)
				Includes += "<tr>"
										"<td align=\"right\"><code>" + *i + "</code> -></td>"
										"<td align=\"right\"><a href=\"pov:" + Inc_Path + "?info\">" + Inc_Path + "</a></td>"
										"</tr>";
			else Includes += "<tr><td align=\"right\"><code>" + *i + "</code></td></tr>";
		}
		if (Includes != QString::null)
		{
			Includes.prepend("<table cellspacing=2 style=\"padding: 0; margin: 0; border: 0;\">");
			Includes.append("</table>");
		}

		// scene properties: version, author etc.
		QString Scene_Properties("<h2>Scene Properties:</h2>");
		Scene_Properties.prepend(Scene_Menu);

		{
			QStringList Col1;
			QStringList Col2;

			if (Scene.povray_version() > 0.0)
			{
				Col1 += i18n("POV-Ray version");
				if (Scene.povray_executable() == "povray")
					Col2 += QString("%1").arg(Scene.povray_version());
				else
					Col2 += QString("%1 %1").arg(Scene.povray_executable()).arg(Scene.povray_version());
			}

			if (Scene.author().length() > 0)
			{
				Col1 += i18n("Scene author:");
				Col2 += Scene.author();
			}

			if (Scene.description().length() > 0)
			{
				Col1 += i18n("Scene description:");
				Col2 += Scene.description();
			}

			if (Scene.options().length() > 0)
			{
				Col1 += i18n("Suggested render options:");
				Col2 += "<code>" + Scene.options() + "</code>";
			}

			if (Includes.length() > 0)
			{
				Col1 += i18n("Included files:");
				Col2 += Includes;
			}

			Col1 += i18n("line count");
			Col2 += QString("%1").arg(Scene.line_count());
			
			Scene_Properties += HTMLInfoTable(Col1, Col2);
		}

		content = content
				.arg(Scene.pov_file())
				.arg(logo)
				.arg(Scene.filename())
				.arg(i18n("file information"))
				.arg(Scene_Properties)
				.arg(KIO_POV_NAME " version " KIO_POV_VERSION);
			
		infoMessage(i18n("Scene information generated."));
		
		mimeType("text/html");
		data(content.local8Bit());
	}
	// ------------------------------------------------------------------------
	/*
	else
	{
		mimeType("text/html");
		data(GenerateInfo(url).local8Bit());
	}
	*/
	
}

/* --------------------------------------------------------------------------- */

/**
process stdout from POV-Ray - write image data to image stream
 */
 
void kio_povProtocol::slotSetDataStdOutput(KProcess*, char *s, int len) 
{
	processedBytes += len;
	processedSize(processedBytes);
	/*
	QByteArray array;
	array.setRawData(s, len);
	data(array);
	array.resetRawData(s, len);
	*/
	Povray_ImageStream->writeRawBytes(s, len);
	if (wasKilled())
	{
		PovrayKProcess->kill();
		delete PovrayKProcess; PovrayKProcess = 0;
		delete Povray_ImageStream; Povray_ImageStream = 0;
	}
}

/* --------------------------------------------------------------------------- */

/**
process stderr from POV-Ray - store and analyse log, update status
if necessary
 */
 
void kio_povProtocol::slotGetStdError(KProcess*, char *s, int len)
{
	QString text = QString::fromLocal8Bit(s, len);
	ErrorStream += text;
	QString message = ErrorStream.right(100);
	int pos = message.findRev( QRegExp("[\\r\\n][ ]*[0-9]+:") );
	if (pos >= 0)
	{
		pos++;
		int end_pos = message.findRev( QRegExp("[\\r\\n]"));
		if (end_pos<=pos) end_pos = message.length();
		message = message.mid(pos,end_pos-pos);
		infoMessage(i18n(QString("POV-Ray: %1").arg(message)));
		pos = message.find("Rendering line");
		if (pos >= 0)
		{
			int pos2 = message.find("of", pos);
			if (pos >= 0)
			{
				long line = message.mid(pos+15,pos2-pos).toULong();
				long line_cnt = message.mid(pos2+2,end_pos-pos2).toULong();
				processedPercent(float(line)/float(line_cnt));
			}
		}
	}
	if (wasKilled())
	{
		PovrayKProcess->kill();
		delete PovrayKProcess; PovrayKProcess = 0;
		if (Povray_ImageStream != 0)
		{
			delete Povray_ImageStream;
			Povray_ImageStream = 0;
		}
	}
}

/* --------------------------------------------------------------------------- */

/**
process stderr from POV-Ray - store and analyse log, update status
if necessary, version for animation renders
 */
 
void kio_povProtocol::slotGetStdErrorAnim(KProcess*, char *s, int len)
{
	QString text = QString::fromLocal8Bit(s, len);
	ErrorStream += text;
	QString message = ErrorStream.right(100);
	int pos = message.findRev( QRegExp("[\\r\\n][ ]*[0-9]+:") );
	if (pos >= 0)
	{
		pos++;
		int end_pos = message.findRev( QRegExp("[\\r\\n]"));
		if (end_pos<=pos) end_pos = message.length();
		message = message.mid(pos,end_pos-pos);
		pos = message.find("Processing Frame");
		if (pos >= 0)
		{
			infoMessage(i18n(QString("POV-Ray: %1").arg(message)));
			int pos2 = message.find("of", pos);
			if (pos >= 0)
			{
				long line = message.mid(pos+15,pos2-pos).toULong();
				long line_cnt = message.mid(pos2+2,end_pos-pos2).toULong();
				processedPercent(float(line)/float(line_cnt));
			}
		}
	}
	/*
	int pos = text.find("Processing Frame");
	if (pos >= 0)
	{
		int end_pos = text.find( QRegExp("[\\r\\n]"), pos );
		if (end_pos<0) end_pos = text.find( QRegExp("$"), pos );
		if (end_pos >= 0)
		{
			QString message = text.mid(pos,end_pos-pos);
			infoMessage(i18n(QString("POV-Ray: %1").arg(message)));
			int pos2 = text.find("of", pos);
			if (pos2 >= 0)
			{
				long line = text.mid(pos+17,pos2-pos).toULong();
				long line_cnt = text.mid(pos2+2,end_pos-pos2).toULong();
				processedPercent(float(line)/float(line_cnt));
			}
		}
	}
	*/
	/*
	int pos = text.find( QRegExp("[\\r\\n][ ]*[0-9]+:") );
	if (pos<0) pos = text.find( QRegExp("^[ ]*[0-9]+:") );
	if (pos >= 0)
	{
		pos++;
		int end_pos = text.find( QRegExp("[\\r\\n]"), pos );
		if (end_pos<0) end_pos = text.find( QRegExp("$"), pos );
		if (end_pos >= 0)
		{
			QString message = text.mid(pos,end_pos-pos);
			pos = text.find("Processing Frame");
			if (pos >= 0)
			{
				infoMessage(i18n(QString("POV-Ray: %1").arg(message)));
				int pos2 = text.find("of", pos);
				if (pos2 >= 0)
				{
					long line = text.mid(pos+17,pos2-pos).toULong();
					long line_cnt = text.mid(pos2+2,end_pos-pos2).toULong();
					processedPercent(float(line)/float(line_cnt));
				}
			}
		}
	}
	*/
	if (wasKilled())
	{
		PovrayKProcess->kill();
		delete PovrayKProcess; PovrayKProcess = 0;
	}
}

/* --------------------------------------------------------------------------- */

/**
process stderr from FFMpeg - store and analyse log, update status if necessary
 */
 
void kio_povProtocol::slotMpegStdOut(KProcess*, char *s, int len)
{
	QString text = QString::fromLocal8Bit(s, len);
	FFMpegErrorStream += text;
	if (wasKilled())
	{
		FFMpegKProcess->kill();
		delete FFMpegKProcess; FFMpegKProcess = 0;
	}
}

int kdemain(int argc, char **argv)
{
	KInstance  instance("kio_pov");

	kdDebug(DEBUG_AREA) << "*** Starting kio_pov " << endl;

	if (argc != 4)
	{
		fprintf(stderr, "Usage: kio_pov protocol domain-socket1 domain-socket2\n");
		exit(-1);
	}

	kio_povProtocol slave(argv[2], argv[3]);
	slave.dispatchLoop();

	kdDebug(DEBUG_AREA) << "*** kio_pov Done" << endl;
	return 0;
}
